/*
 * TreeFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "TreeFromStringParser.h"
#include "exception/AlibException.h"
#include "tree/TreeClasses.h"

#include "../StringApi.hpp"
#include "alphabet/SubtreeWildcardSymbol.h"

namespace tree {

Tree TreeFromStringParser::parseTree ( std::istream & input ) const {
	return parseTree ( input, std::set < FEATURES > ( { FEATURES::RANKED_TREE, FEATURES::RANKED_PATTERN, FEATURES::UNRANKED_TREE, FEATURES::UNRANKED_PATTERN } ) );
}

Tree TreeFromStringParser::parseTree ( std::istream & input, const std::set < FEATURES > & features ) const {
	TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
			return Tree ( UnrankedPattern ( subtreeWildcard, UnrankedNode ( subtreeWildcard, { } ) ) );
		} else {
			alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
			return Tree ( RankedPattern ( subtreeWildcard, RankedNode ( subtreeWildcard, { } ) ) );
		}
	} else {
		m_TreeLexer.putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;

		token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK ) {
			rank = std::stou ( token.value );

			std::vector < tree::RankedNode * > childs;
			bool isPattern = false;

			for ( unsigned i = 0; i < rank; i++ )
				childs.push_back ( parseRankedContent ( input, isPattern ) );

			if ( isPattern ) {
				alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
				RankedPattern tree ( std::move ( subtreeWildcard ), RankedNode ( alphabet::RankedSymbol ( symbol, rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_PATTERN ) ) return Tree ( tree );
			} else {
				RankedTree tree ( RankedNode ( alphabet::RankedSymbol ( symbol, rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_TREE ) ) return Tree ( tree );
			}
		} else {
			std::vector < tree::UnrankedNode * > childs;
			bool isPattern = false;

			while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
				m_TreeLexer.putback ( input, token );
				childs.push_back ( parseUnrankedContent ( input, isPattern ) );
				rank++;
				token = m_TreeLexer.next ( input );
			}

			if ( token.type != TreeFromStringLexer::TokenType::BAR )
				throw exception::AlibException ( "Missing bar" );

			if ( isPattern ) {
				alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
				UnrankedPattern tree ( std::move ( subtreeWildcard ), UnrankedNode ( symbol, childs ) );

				if ( features.count ( FEATURES::UNRANKED_PATTERN ) ) return Tree ( tree );
			} else {
				UnrankedTree tree ( UnrankedNode ( symbol, childs ) );

				if ( features.count ( FEATURES::UNRANKED_TREE ) ) return Tree ( tree );
			}
		}

		throw exception::AlibException ( "Invalid input" );
	}
}

tree::RankedNode * TreeFromStringParser::parseRankedContent ( std::istream & input, bool & isPattern ) const {
	TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		isPattern = true;
		alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
		return new RankedNode ( std::move ( subtreeWildcard ), { } );
	} else {
		m_TreeLexer.putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;
		std::vector < tree::RankedNode * > childs;

		TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK )
			rank = std::stou ( token.value );
		else
			throw exception::AlibException ( "Missing rank" );

		for ( unsigned i = 0; i < rank; i++ )
			childs.push_back ( parseRankedContent ( input, isPattern ) );

		return new RankedNode ( alphabet::RankedSymbol ( std::move ( symbol ), rank ), std::move ( childs ) );
	}
}

tree::UnrankedNode * TreeFromStringParser::parseUnrankedContent ( std::istream & input, bool & isPattern ) const {
	TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = m_TreeLexer.next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::AlibException ( "Missing bar" );

		isPattern = true;
		alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
		return new UnrankedNode ( std::move ( subtreeWildcard ), { } );
	} else {
		m_TreeLexer.putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;
		std::vector < tree::UnrankedNode * > childs;

		TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

		while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
			m_TreeLexer.putback ( input, token );
			childs.push_back ( parseUnrankedContent ( input, isPattern ) );
			rank++;
			token = m_TreeLexer.next ( input );
		}

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::AlibException ( "Missing bar" );

		return new UnrankedNode ( std::move ( symbol ), std::move ( childs ) );
	}
}

} /* namespace tree */
